;; Demonstration of parallel list structures - one proc per item

(*DEFVAR LTYP (!! 'FREE))   ; Type of item, one of FREE CONS SYM NUM
(*DEFVAR HD   (!! 0))       ; Head (pointer)
(*DEFVAR TL   (!! 0))       ; Tail (pointer)
(*DEFVAR NM   (!! 0))       ; Numeric value
(*DEFVAR SYM  (!! 'NONE))   ; Symbolic value

(*DEFVAR REFC (!! 0))       ; Reference count

(defvar roots nil)          ; List of root objects in Cm (cube addresses)

(defun setup ()
"Setup empty memory"
 (*set LTYP (!! 'FREE)) (*set REFC (!! 0))
 (setf (pref LTYP 0) 'SYM)    ; Install NIL at 0
 (setf (pref SYM 0) NIL)  (setf (pref REFC 0) 1)
 (setq roots '(0))
)

;; Install list structure, preserving sharing within L - returns addr of rep.
;; Maintains full sharing of numbers and symbols

(defun install (L)
 (let (r)
  (*when (or!! (eq!! LTYP (!! 'FREE)) (eq!! LTYP (!! 'CONS)))
      (*set SYM (!! 'NONE)))
  (+f (setq r (installx L)))
  (push r roots)
  r)
)

(defun installx (L)
 (let (r)
   (cond ((consp L)
	  (setq r (cpresent L 'CONS)) (if r r (mkcons L)))
	 ((symbolp L)
	  (setq r (cpresent L 'SYM)) (if r r (mksym L)))
         ((numberp L)
	  (setq r (npresent L)) (if r r (mknum L)))
	 (T (error "Unexpected object~%~S" L)))
  )
)

;; Check for presence of list structure L of type ty (CONS or SYM) in Cm
;; Returns cube address if present, nil otherwise

(defun cpresent (L ty)
  (*when (and!! (eq!! LTYP (!! ty)) (eq!! SYM (!! L)))
      (*max (self-address!!)))
)

;; Check for presence of number N in Cm
;; Returns cube address if present, nil otherwise

(defun npresent (N)
  (*when (and!! (eq!! LTYP (!! 'NUM)) (=!! NM (!! N)))
      (*max (self-address!!)))
)

;; Construct list structure when L not already present

(defun mkcons (L)
 (let ((c (newp)) d)
   (setf (pref LTYP c) 'CONS)
   (setf (pref SYM c) L)
   (setf (pref HD c) (setq d (installx (car L)))) (+f d)
   (setf (pref TL c) (setq d (installx (cdr L)))) (+f d)
   c)
)

;; Construct new symbol

(defun mksym (S)
 (let ((c (newp)))  (setf (pref LTYP c) 'SYM)  (setf (pref SYM c) S)  c)
)

;; Construct new number

(defun mknum (N)
 (let ((c (newp)))  (setf (pref LTYP c) 'NUM)  (setf (pref NM c) N)  c)
)

;; Allocate free processor if possible.  Error if not.

(defun newp ()
 (let (c)
   (*when (eq!! LTYP (!! 'FREE)) (setq c (*min (self-address!!))))
   (if (not c) (error "Out of processors~%"))
   c)
)

;; Delete p-list at cube address c

(defun delp (c)
  (when (member c roots) (-f c)
    (setq roots (nset-difference roots (list c))) T)
)

;; Increment one ref count; NIL (proc 0) is ignored

(defun +f (c)
 (when (> c 0) (setf (pref REFC c) (1+ (pref REFC c))))
)

;; Decrement one ref count

(defun -f (c)
 (when (> c 0) (setf (pref REFC c) (1- (pref REFC c))))
)

;; Adjust reference counts in parallel.  All cells pointed to by P
;; in active processors are adjusted by d (scalar integer)

(*defun F!! (P d)
 (*let (W)
   (*all (*set W (!! 0)))
   (*when (plusp!! P) (*pset :add (!! d) W P))
   (*all (*set REFC (+!! REFC W)))
  )
)

;; Convenient macros for parallel refc adjustment

(defmacro +F!! (P) `(F!! ,P  1))
(defmacro -F!! (P) `(F!! ,P -1))

;; Garbage collection, n cycles

(defun rfgc (&optional (n 1))
 (do ((i 0 (incf i)) (more t))
     ((not (and more (< i n))))
     (*when (and!! (not!! (eq!! LTYP (!! 'FREE))) (zerop!! REFC))
	(*when (eq!! LTYP (!! 'CONS))
	   (when (setq more (*or T!!)) (F!! HD -1) (F!! TL -1)))
	(*set LTYP (!! 'FREE)))
   )  ; end do
 (*when (eq!! LTYP (!! 'FREE)) (*set REFC (!! 0)))
)

;; Extract list structure at c, retaining structure sharing.
;; Tolerates circular structures

(defun xtr (c)
 (*when (eq!! LTYP (!! 'CONS)) (*set SYM NIL!!))
 (xtrx c)
)

(defun xtrx (c)
 (let ((ty (pref LTYP c)) r)
   (cond ((eq ty 'CONS)
	  (setq r (pref SYM c))
	  (if r r
	    (progn (setq r (setf (pref SYM c) (cons NIL NIL)))
		   (rplaca r (xtrx (pref HD c)))
		   (rplacd r (xtrx (pref TL c))))))
	 ((eq ty 'sym) (pref SYM c))
	 (T (pref NM c)))
  )
)

;; Display non-free portions of memory

(defun dsp ()
  (format t " Addr   Ltyp   Refc    Hd     Tl    Numb    Sym~%")
  (format t "------ ------ ------ ------ ------ ------ -------~%")
  (*when (not!! (eq!! LTYP (!! 'FREE)))
     (do-for-selected-processors (i)
       (format t "~6D ~6A ~6D ~6D ~6D ~6D "
	  i (pref LTYP i) (pref REFC i) (pref HD i) (pref TL i)
	    (pref NM i) )
       (write (pref SYM i) :circle T :level 2 :length 3)
       (terpri)))
  (values)
)